﻿' K8055 Class for Visual Basic .NET
'    Copyright (c) 2011-2017   Vanderhaegen Bart
'    Class Version 2.0
'
'
' I have created this class to provide an easy access to the Velleman K8055 card.
' This class is provided as it is, you should use it at your own risk.
'
' I advise you to use this class as an instant variable of your form. Doing so you
' can access the object of this class from different subs/functions in your form
' class. This is the minimal code you should use to get the card working:
'
'       Public Class frmMain
'         Private card1 As K8055
'
'         Private Sub frmMain_Load() Handles MyBase.Load
'               card1 = new K8055()
'               card1.connect(0)
'         End Sub
'       End Class
'
'
' The methods in this class are named slightly different than the ones in the 
' K8055D.DLL. This shouldn't be a big problem. The functions you can use are
' Explained on by website http://www.bvsystems.be/k8055class2.php



' Enforce Explicit & Strict modes
Option Explicit On
Option Strict On


Public Class K8055

#Region "Hardware Specifications"
    ' Constants with information about the K8055 card. Do NOT change these values as they 
    ' will make your program crash.
    Public Const MAX_ANALOG_OUTPUTS As Integer = 2
    Public Const MAX_ANALOG_INPUTS As Integer = 2
    Public Const MAX_DIGITAL_OUTPUTS As Integer = 8
    Public Const MAX_DIGITAL_INPUTS As Integer = 5
    Public Const MAX_NUMBER_OF_CARDS As Integer = 4
    Public Const MAX_NUMBER_OF_COUNTERS As Integer = 2
    Public Const MAX_ADC_STEPS As Integer = 255
    Public Const MAX_DAC_STEPS As Integer = 255

    ''' <summary>
    ''' Enumeration type for the diffent types of K8055 I/O
    ''' </summary>
    ''' <remarks></remarks>
    Private Enum channelType
        DIGITAL_INPUT
        DIGITAL_OUTPUT
        ANALOG_INPUT
        ANALOG_OUTPUT
    End Enum
#End Region



#Region "DLL Declarations"
    ' Initialisation & Connection
    Private Declare Function OpenDevice Lib "k8055d.dll" (ByVal CardAddress As Integer) As Integer
    Private Declare Sub CloseDevice Lib "k8055d.dll" ()
    Private Declare Function Version Lib "k8055d.dll" () As Integer
    Private Declare Function SearchDevices Lib "k8055d.dll" () As Integer
    Private Declare Function SetCurrentDevice Lib "k8055d.dll" (ByVal CardAddress As Integer) As Integer

    ' Analog Channels
    Private Declare Function ReadAnalogChannel Lib "k8055d.dll" (ByVal Channel As Integer) As Integer
    Private Declare Sub ReadAllAnalog Lib "k8055d.dll" (ByRef Data1 As Integer, ByRef Data2 As Integer)
    Private Declare Sub OutputAnalogChannel Lib "k8055d.dll" (ByVal Channel As Integer, ByVal Data As Integer)
    Private Declare Sub OutputAllAnalog Lib "k8055d.dll" (ByVal Data1 As Integer, ByVal Data2 As Integer)
    Private Declare Sub ClearAnalogChannel Lib "k8055d.dll" (ByVal Channel As Integer)
    Private Declare Sub SetAllAnalog Lib "k8055d.dll" ()
    Private Declare Sub ClearAllAnalog Lib "k8055d.dll" ()
    Private Declare Sub SetAnalogChannel Lib "k8055d.dll" (ByVal Channel As Integer)

    ' Digital Output Channels
    Private Declare Sub WriteAllDigital Lib "k8055d.dll" (ByVal Data As Integer)
    Private Declare Sub ClearDigitalChannel Lib "k8055d.dll" (ByVal Channel As Integer)
    Private Declare Sub ClearAllDigital Lib "k8055d.dll" ()
    Private Declare Sub SetDigitalChannel Lib "k8055d.dll" (ByVal Channel As Integer)
    Private Declare Sub SetAllDigital Lib "k8055d.dll" ()

    ' Digital Input channels
    Private Declare Function ReadDigitalChannel Lib "k8055d.dll" (ByVal Channel As Integer) As Boolean
    Private Declare Function ReadAllDigital Lib "k8055d.dll" () As Integer

    ' Counters
    Private Declare Function ReadCounter Lib "k8055d.dll" (ByVal CounterNr As Integer) As Integer
    Private Declare Sub ResetCounter Lib "k8055d.dll" (ByVal CounterNr As Integer)
    Private Declare Sub SetCounterDebounceTime Lib "k8055d.dll" (ByVal CounterNr As Integer, ByVal DebounceTime As Integer)
#End Region


#Region "Events & Declarations"
    ' Defined events
    Public Event AnalogInputChanged(cardid As Integer, channel_of_card As Integer, channel As Integer, value As Integer)
    Public Event digitalInputChanged(cardid As Integer, channel_of_card As Integer, channel As Integer, state As Boolean)
    Public Event digitalInputChangedOn(cardid As Integer, channel_of_card As Integer, channel As Integer, state As Boolean)
    Public Event digitalInputChangedOff(cardid As Integer, channel_of_card As Integer, channel As Integer, state As Boolean)
    Public Event readCycleCompleted()

    ' State variables. They contain the cached state of the Input Channels
    Private digitalOutputChannels(MAX_NUMBER_OF_CARDS * MAX_DIGITAL_OUTPUTS - 1) As Boolean
    Private digitalInputChannels(MAX_NUMBER_OF_CARDS * MAX_DIGITAL_INPUTS - 1) As Boolean
    Private analogOutputChannels(MAX_NUMBER_OF_CARDS * MAX_ANALOG_OUTPUTS - 1) As Integer
    Private analogInputChannels(MAX_NUMBER_OF_CARDS * MAX_ANALOG_INPUTS - 1) As Integer
    Private connected_cards_array(MAX_NUMBER_OF_CARDS - 1) As Boolean

    ' Define a Timer that will be checked in a regular Interval
    Private WithEvents readTimer As New System.Windows.Forms.Timer()
    Private _timerEnabled As Boolean = False
    Private _timerInterval As Integer = 100
    Private _cacheEnabled As Boolean = True
    Private _initState As Boolean = False
    Private activeCard As Integer = -1

    ' Internal variables
    Private _use_DAC_emulation As Boolean
    Private Shared Property currentInstance As K8055
    Private counted_IO_operations As ULong

    ' Defined the allowed polling interval (ms)
    Private Const MIN_POLLING_INTERVAL As Integer = 1
    Private Const MAX_POLLING_INTERVAL As Integer = 60000

    ' Event configuration word
    Private currentEventConfiguration As EventConfiguration

    ''' <summary>
    ''' Possible event configuration words. It is possible to combine multiple 
    ''' configurations (ALL_ANALOG + ALL_DIGITAL) ...
    ''' </summary>
    ''' <remarks></remarks>
    Public Enum EventConfiguration
        ALL_DIGITAL = 1
        ALL_ANALOG = 2
        CARD_0_ANALOG = 4
        CARD_0_DIGITAL = 8
        CARD_1_ANALOG = 16
        CARD_1_DIGITAL = 32
        CARD_2_ANALOG = 64
        CARD_2_DIGITAL = 128
        CARD_3_ANALOG = 256
        CARD_3_DIGITAL = 512
    End Enum
#End Region




#Region "Properties"
    ''' <summary>
    ''' When set an exception will be raised when trying to obtain a value from a card that is not connected.
    ''' </summary>
    ''' <value></value>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Property throwExceptionOnMissingCard As Boolean

    ''' <summary>
    ''' Enable the support for multiple Interface cards
    ''' </summary>
    ''' <value></value>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Property multiCardMode As Boolean

    ''' <summary>
    ''' Specifies that, if you are using a single K8055 card, its is always threated as card address 0. With
    ''' digital outputs 1-8, digital inputs 1-5, analog inputs 1-2, analog outputs 1-2.
    ''' </summary>
    ''' <value>The desired option (default = activated)</value>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Property remap_single_card_as_first As Boolean = True

    ''' <summary>
    ''' Emulate the analog output channels as DIGITAL OUTPUTS. If enabled they can be adressed through 
    ''' digital output 9 and 10 of each card.
    ''' </summary>
    ''' <value></value>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Property emulate_DAC_as_digital_outputs As Boolean
        Set(value As Boolean)
            _use_DAC_emulation = value
            convertDigitalOutputArray()
        End Set
        Get
            Return _use_DAC_emulation
        End Get
    End Property

    ''' <summary>
    ''' Enables or disables the built-in polling timer. When the timer is disabled, no Events can be raised.
    ''' </summary>
    ''' <value>The desired state of the timer</value>
    ''' <returns>The current state of the timer</returns>
    ''' <remarks></remarks>
    Private Property enable_events() As Boolean
        Get
            Return _timerEnabled
        End Get
        Set(value As Boolean)
            _timerEnabled = value
            If Not _timerEnabled Then _cacheEnabled = False
        End Set
    End Property

    ''' <summary>
    ''' Sets the polling interval that is used to scan the K8055
    ''' </summary>
    ''' <value>The desired interval in milliseconds</value>
    ''' <returns>The current timer interval in milliseconds</returns>
    ''' <remarks></remarks>
    Private Property events_polling_interval() As Integer
        Get
            Return _timerInterval
        End Get
        Set(value As Integer)
            If isBetween(value, MIN_POLLING_INTERVAL, MAX_POLLING_INTERVAL) Then
                _timerInterval = value
                readTimer.Interval = _timerInterval
            Else
                Throw New ArgumentException(String.Format("The interval for the event polling timer must be between {0} and {1} ms",
                                            MIN_POLLING_INTERVAL,
                                            MAX_POLLING_INTERVAL))
            End If
        End Set
    End Property

    ''' <summary>
    ''' Returns a boolean to indicate if any K8055 is currently connected
    ''' </summary>
    ''' <returns>True if the K8055 is currently initialized</returns>
    ''' <remarks></remarks>
    Public ReadOnly Property isConnected() As Boolean
        Get
            Return countConnectedCards() > 0
        End Get
    End Property

    ''' <summary>
    ''' Returns a boolean to indicate if the specified K8055 is currently connected
    ''' </summary>
    ''' <returns>True if the selected K8055 is currently initialized</returns>
    ''' <remarks></remarks>
    Public ReadOnly Property isConnected(card As Integer) As Boolean
        Get
            If isBetween(card, 0, MAX_NUMBER_OF_CARDS - 1) Then
                Return connected_cards_array(card)
            Else
                Throw New Exception("The specified card number is invalid")
            End If
        End Get
    End Property

    ''' <summary>
    ''' Sets the state of the internal input caching. The recommended value is ENABLE.
    ''' </summary>
    ''' <value>The desired state of the input caching</value>
    ''' <returns>The current state of the input caching</returns>
    ''' <remarks></remarks>
    Public Property cacheEnabled() As Boolean
        Get
            Return _cacheEnabled
        End Get
        Set(value As Boolean)
            _cacheEnabled = value
        End Set
    End Property

    ''' <summary>
    ''' Returns the number of K8055 hardware I/O calls since the creation of the K8055 object
    ''' </summary>
    ''' <value></value>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public ReadOnly Property counted_Hardware_Operations As ULong
        Get
            Return counted_IO_operations
        End Get
    End Property

    ''' <summary>
    ''' Setups the internal event checking mechanism
    ''' </summary>
    ''' <param name="enable">Desired global state</param>
    ''' <param name="polling_interval">The interval to scan the K8055's (50 ms)</param>
    ''' <param name="mode">The configuration word (see K8055.EventConfiguration Enumeration)</param>
    ''' <remarks></remarks>
    Public Sub setupEvents(enable As Boolean, polling_interval As Integer, mode As EventConfiguration)
        events_polling_interval = polling_interval
        enable_events = enable
        currentEventConfiguration = mode
    End Sub
#End Region





#Region "Digital Outputs"
    ''' <summary>
    ''' Enables the specified digital output channel
    ''' </summary>
    ''' <param name="channel">The channel to enable (1-32)</param>
    ''' <remarks></remarks>
    Public Sub setDigitalOutput(channel As Integer)
        Dim card, channel_of_card As Integer
        getChannelInformation(channel, channelType.DIGITAL_OUTPUT, card, channel_of_card)

        If activateCard(card) Then
            digitalOutputChannels(channel - 1) = True
            setDigitalChannelWrapper(channel_of_card)
        End If
    End Sub

    ''' <summary>
    ''' Disables the specified digital output channel
    ''' </summary>
    ''' <param name="channel">The channel to deactivate (1-32)</param>
    ''' <remarks></remarks>
    Public Sub clearDigitalOutput(channel As Integer)
        Dim card, channel_of_card As Integer
        getChannelInformation(channel, channelType.DIGITAL_OUTPUT, card, channel_of_card)
        If activateCard(card) Then
            digitalOutputChannels(channel - 1) = False
            clearDigitalChannelWrapper(channel_of_card)
        End If
    End Sub

    ''' <summary>
    ''' Get the state of the specified digital output channel
    ''' </summary>
    ''' <param name="channel">The channel to read (1-32)</param>
    ''' <returns>True if the output is currently enabled</returns>
    ''' <remarks></remarks>
    Public Function getDigitalOutput(channel As Integer) As Boolean
        ' ByRef parameters are not used - just using this function for input validation
        Dim card, channel_of_card As Integer
        getChannelInformation(channel, channelType.DIGITAL_OUTPUT, card, channel_of_card)
        Return digitalOutputChannels(channel - 1)
    End Function

    ''' <summary>
    ''' Send the selected digital effect - stored as integer value - to the specified card
    ''' </summary>
    ''' <param name="destination_card">The card to send the effect to</param>
    ''' <param name="effect">Integer value containing the desired effect (the rightmost bit 
    ''' contains the value for output 1). If the emulate_DAC_as_digital_outputs is enabled, 
    ''' bit 9 and 10 will set the analog outputs.</param>
    ''' <remarks></remarks>
    Public Sub setDigitalEffect(destination_card As Integer, effect As Integer)
        ' Check if remapping is active, remap if necessary
        If remap_single_card_as_first Then
            ' If only one card is in use, the destination_card setting may be overwritten
            If countConnectedCards() = 1 Then
                For i As Integer = 0 To connected_cards_array.Count - 1
                    If connected_cards_array(i) Then
                        destination_card = i
                    End If
                Next
            End If
        End If

        ' Activate the card if this wasn't done yet
        If activeCard <> destination_card Then
            activateCard(destination_card)
        End If

        ' Check if the card is using either 8 or 8+2 (emulated) Digital Outputs
        Dim outputsPerCard As Integer
        If _use_DAC_emulation Then
            outputsPerCard = MAX_DIGITAL_OUTPUTS + MAX_ANALOG_OUTPUTS
        Else
            outputsPerCard = MAX_DIGITAL_OUTPUTS
        End If

        Dim firstBits As Integer = effect And 255
        increaseIoOperations()
        WriteAllDigital(firstBits)

        If emulate_DAC_as_digital_outputs Then
            Dim lastBits As Integer = (effect And 256 + 512) >> 8
            Dim value1, value2 As Integer

            If (lastBits And 1) > 0 Then
                value1 = 255
            End If

            If (lastBits And 2) > 0 Then
                value2 = 255
            End If

            increaseIoOperations()
            OutputAllAnalog(value1, value2)
        End If
    End Sub

    ''' <summary>
    ''' Sets all digital outputs
    ''' </summary>
    ''' <remarks></remarks>
    Public Sub setAllDigitalOutputs()
        For i As Integer = 1 To MAX_NUMBER_OF_CARDS
            If activateCard(i - 1) Then
                increaseIoOperations()
                SetAllDigital()

                If _use_DAC_emulation Then
                    For j As Integer = 1 To MAX_ANALOG_OUTPUTS
                        increaseIoOperations()
                        SetAnalogChannel(j)
                    Next
                End If

                ' Update the Array
                For j As Integer = 1 To MAX_DIGITAL_OUTPUTS
                    digitalOutputChannels(i * MAX_NUMBER_OF_CARDS + j - 2) = True
                Next
            End If
        Next
    End Sub

    ''' <summary>
    '''  Sets all digital outputs of the specified card
    ''' </summary>
    ''' <param name="card">The hardware address of the card (0-3)</param>
    ''' <remarks></remarks>
    Public Sub setAllDigitalOutputs(card As Integer)
        If isBetween(card, 0, MAX_NUMBER_OF_CARDS - 1) Then
            For i As Integer = 1 To 8
                setDigitalOutput(card * MAX_DIGITAL_OUTPUTS + i)
            Next
        End If
    End Sub

    ''' <summary>
    '''  Sets all digital outputs of the specified card
    ''' </summary>
    ''' <param name="card">The hardware address of the card (0-3)</param>
    ''' <remarks></remarks>
    Public Sub clearAllDigitalOutputs(card As Integer)
        If isBetween(card, 0, MAX_NUMBER_OF_CARDS - 1) Then
            For i As Integer = 1 To 8
                clearDigitalOutput(card * MAX_DIGITAL_OUTPUTS + i)
            Next
        End If
    End Sub

    ''' <summary>
    '''  Sets all digital outputs of the specified card
    ''' </summary>
    ''' <param name="card">The hardware address of the card (0-3)</param>
    ''' <remarks></remarks>
    Public Sub toggleAllDigitalOutputs(card As Integer)
        If isBetween(card, 0, MAX_NUMBER_OF_CARDS - 1) Then
            For i As Integer = 1 To 8
                toggleDigitalOutput(card * MAX_DIGITAL_OUTPUTS + i)
            Next
        End If
    End Sub
    ''' <summary>
    ''' Clears all digital outputs
    ''' </summary>
    ''' <remarks></remarks>
    Public Sub clearAllDigitalOutputs()
        For i As Integer = 1 To MAX_NUMBER_OF_CARDS
            If activateCard(i - 1) Then
                increaseIoOperations()
                ClearAllDigital()

                If _use_DAC_emulation Then
                    increaseIoOperations()
                    OutputAllAnalog(0, 0)
                End If

                For j As Integer = 1 To MAX_DIGITAL_OUTPUTS
                    digitalOutputChannels(i * MAX_NUMBER_OF_CARDS + j - 2) = False
                Next
            End If
        Next
    End Sub

    ''' <summary>
    ''' Toggles the state of a digital output channel
    ''' </summary>
    ''' <remarks></remarks>
    Public Sub toggleAllDigitalOutputs()
        For i As Integer = 1 To MAX_NUMBER_OF_CARDS * MAX_DIGITAL_OUTPUTS
            toggleDigitalOutput(i)
        Next
    End Sub

    ''' <summary>
    ''' Toggles the state of a digital input channel
    ''' </summary>
    ''' <param name="channel">The channel to toggle (1-32)</param>
    ''' <remarks></remarks>
    Public Sub toggleDigitalOutput(channel As Integer)
        If getDigitalOutput(channel) Then
            clearDigitalOutput(channel)
        Else
            setDigitalOutput(channel)
        End If
    End Sub
#End Region




#Region "Digital Input Channels"
    ''' <summary>
    ''' Get the value of the specified input channel
    ''' </summary>
    ''' <param name="channel">The channel to poll (1-20)</param>
    ''' <returns>True if the specified channel is currently active</returns>
    ''' <remarks></remarks>
    Public Function getDigitalInput(channel As Integer) As Boolean
        If isBetween(channel, 1, MAX_DIGITAL_INPUTS) Then
            If _timerEnabled And _cacheEnabled And Not _initState Then
                ' Cache mode enabled
                Return digitalInputChannels(channel - 1)
            Else

                Dim card, channel_of_card As Integer
                getChannelInformation(channel, channelType.DIGITAL_OUTPUT, card, channel_of_card)

                If activateCard(card) Then
                    increaseIoOperations()
                    Return ReadDigitalChannel(channel_of_card)
                Else
                    ' This can never happen
                    Return False
                End If
            End If
        Else
            Throw New ArgumentOutOfRangeException(
                String.Format("Incorrect digital channel specified: Specify a channel between 1 and {0}", MAX_DIGITAL_INPUTS))
        End If
    End Function

    ''' <summary>
    ''' Get the value of all input channels
    ''' </summary>
    ''' <returns>An Integer value with the last four bits representing the digital inputs</returns>
    ''' <remarks></remarks>
    Public Function getAllDigitalInputs() As Integer
        If _timerEnabled And _cacheEnabled And Not _initState Then
            ' Cache mode enabled
            Dim sum As Integer
            For i As Integer = 1 To MAX_DIGITAL_INPUTS
                If digitalInputChannels(i - 1) Then
                    sum += CInt(Math.Pow(2, i - 1))
                End If
            Next
            Return sum
        Else
            ' Cache mode disabled
            increaseIoOperations()
            Return ReadAllDigital
        End If
    End Function

    ''' <summary>
    ''' Get the values of all Digital Input channels as an Array of Boolean values
    ''' </summary>
    ''' <returns>An array of boolean values. Warning: the returned array is ONE-based (element 0 
    ''' exists but is not being used: there are 2 channels but 3 allocations in the array)</returns>
    ''' <remarks></remarks>
    Public Function getAllDigitalInputsAsArray() As Boolean()
        Dim results(MAX_NUMBER_OF_CARDS * MAX_DIGITAL_INPUTS) As Boolean   ' Actual number of allocated array items is MAX_DIGITAL_INPUTS + 1

        For currentCard As Integer = 0 To MAX_NUMBER_OF_CARDS - 1
            If connected_cards_array(currentCard) Then
                activateCard(currentCard)

                Dim values As Integer = getAllDigitalInputs()
                For i As Integer = 1 To MAX_DIGITAL_INPUTS
                    results(currentCard * MAX_DIGITAL_INPUTS + i) = (values And CInt(Math.Pow(2, i - 1))) > 0
                Next i
            End If
        Next

        Return results
    End Function

#End Region




#Region "Counters"
    ''' <summary>
    ''' Read the value of the hardware counter
    ''' </summary>
    ''' <param name="number">The counter to read</param>
    ''' <returns>The current number hold by the counter</returns>
    ''' <remarks></remarks>
    Public Function readCounter_(number As Integer) As Integer
        If isBetween(number, 1, MAX_NUMBER_OF_COUNTERS) Then
            Return ReadCounter(number)
        Else
            Throw New Exception("You specified an non-existing counter number")
        End If
    End Function

    ''' <summary>
    ''' Reset the specified hardware counter
    ''' </summary>
    ''' <param name="number">The counter to reset</param>
    ''' <remarks></remarks>
    Public Sub resetCounter_(number As Integer)
        If isBetween(number, 1, MAX_NUMBER_OF_COUNTERS) Then
            ResetCounter(number)
        Else
            Throw New Exception("You specified an non-existing counter number")
        End If
    End Sub

    ''' <summary>
    ''' Set the counter debounce time
    ''' </summary>
    ''' <param name="number">Number of the counter</param>
    ''' <param name="debounceTime">Desired debounce time</param>
    ''' <remarks></remarks>
    Public Sub setCounterDebounceTime_(number As Integer, debounceTime As Integer)
        If isBetween(number, 1, MAX_NUMBER_OF_COUNTERS) Then
            SetCounterDebounceTime(number, debounceTime)
        Else
            Throw New Exception("You specified an non-existing counter number")
        End If


    End Sub
#End Region




#Region "Connecting K8055"
    ''' <summary>
    ''' Initialises the K8055 after creating a connection
    ''' </summary>
    ''' <remarks></remarks>
    Private Sub initializeConnection()
        ' Clear any analog outputs.
        clearAllAnalogOutputs()
        clearAllDigitalOutputs()
        _initState = True
    End Sub

    ''' <summary>
    ''' Connect to a K8055 by specifying the hardware address as an Integer number
    ''' </summary>
    ''' <param name="cardAddress">An integer number containing the hardware address</param>
    ''' <returns>True if the connection attempt was successfully</returns>
    ''' <remarks></remarks>
    Public Function connect(cardAddress As Integer) As Boolean
        If isBetween(cardAddress, 0, 3) Then
            Return connectDevice(cardAddress)
        Else
            Throw New ArgumentOutOfRangeException("Incorrect card address specified: Specify an adress between 0 and 3.")
        End If
    End Function

    ''' <summary>
    ''' Connect to a K8055 by specifying the hardware address as booleans indicating 
    ''' the states of jumpers SK5 and SK6
    ''' </summary>
    ''' <param name="SK5">The value of SK5</param>
    ''' <param name="SK6">The value of SK6</param>
    ''' <returns>True if the connection attempt was successfully</returns>
    ''' <remarks></remarks>
    Public Function connect(SK5 As Boolean, SK6 As Boolean) As Boolean
        Dim cardAddress As Integer = 3
        If SK5 Then cardAddress -= 1
        If SK6 Then cardAddress -= 2
        Return connectDevice(cardAddress)
    End Function

    ''' <summary>
    ''' Calculates the card number and channel_of_card of the supplied channel number. Information is given
    ''' back by ByRef variables.
    ''' </summary>
    ''' <param name="channel">The general channel to convert (1-8, 9-10 if dac_emulation is enabled)</param>
    ''' <param name="typeOfChannel">Type of channel (ANALOG, DIGITAL, ...)</param>
    ''' <param name="card">The calculated card</param>
    ''' <param name="channel_of_card">The calculated channel of the card</param>
    ''' <remarks></remarks>
    Private Sub getChannelInformation(ByVal channel As Integer, ByVal typeOfChannel As channelType, ByRef card As Integer, ByRef channel_of_card As Integer)
        Dim totalChannelsPerCard As Integer

        Select Case typeOfChannel
            Case channelType.ANALOG_INPUT
                totalChannelsPerCard = MAX_ANALOG_INPUTS
            Case channelType.ANALOG_OUTPUT
                totalChannelsPerCard = MAX_ANALOG_OUTPUTS
            Case channelType.DIGITAL_INPUT
                totalChannelsPerCard = MAX_DIGITAL_INPUTS
            Case channelType.DIGITAL_OUTPUT
                If _use_DAC_emulation Then
                    totalChannelsPerCard = MAX_DIGITAL_OUTPUTS + MAX_ANALOG_OUTPUTS
                Else
                    totalChannelsPerCard = MAX_DIGITAL_OUTPUTS
                End If
            Case Else
                ' nothing
        End Select

        If isBetween(channel, 1, totalChannelsPerCard * MAX_NUMBER_OF_CARDS) Then
            card = (channel - 1) \ totalChannelsPerCard
            channel_of_card = 1 + ((channel - 1) Mod totalChannelsPerCard)
        Else
            Throw New ArgumentOutOfRangeException(String.Format("The specified channel does not exist! Selct a channel between 1 and {0}", totalChannelsPerCard * MAX_NUMBER_OF_CARDS))
        End If
    End Sub

    ''' <summary>
    ''' Activate the specified card
    ''' </summary>
    ''' <param name="number">The card to activate (0-3)</param>
    ''' <returns>Specifies if the operation succceeded</returns>
    ''' <remarks></remarks>
    Public Function activateCard(number As Integer) As Boolean
        ' Remapping function: if only one card is connected, always use this number
        If remap_single_card_as_first Then
            If countConnectedCards() = 1 Then
                ' We only have one card
                Dim activeCard As Integer = -1
                For i As Integer = 0 To MAX_NUMBER_OF_CARDS - 1
                    If connected_cards_array(i) Then
                        number = i
                    End If
                Next
            End If
        End If

        If number <> activeCard Then
            increaseIoOperations()
            Dim h As Integer = SetCurrentDevice(number)
            If h > -1 Then
                activeCard = number
                Return True
            Else
                If throwExceptionOnMissingCard Then
                    Throw New Exception(String.Format("The K8055 with number {0} is currently not connected", number))
                Else
                    Return False
                End If
            End If
        Else
            Return True
        End If
    End Function

    ''' <summary>
    ''' Returns the number of connected K8055 cards()
    ''' </summary>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Function countConnectedCards() As Integer
        Dim numberOfCards As Integer

        For i As Integer = 0 To UBound(connected_cards_array)
            If connected_cards_array(i) Then
                numberOfCards += 1
            End If
        Next

        Return numberOfCards
    End Function

    ''' <summary>
    ''' Private function that executes the connection. PLease use the CONNECT() function.
    ''' </summary>
    ''' <param name="CardAddress">The hardware address</param>
    ''' <returns>True if the connection attempt was successfully</returns>
    ''' <remarks></remarks>
    Private Function connectDevice(CardAddress As Integer) As Boolean
        ' Note: Validating the card address was already done by the calling functions
        If Not multiCardMode And countConnectedCards() > 0 And Not connected_cards_array(CardAddress) Then
            Throw New Exception("There is already a K8055 card connected. Please enable MULTICARDMODE if you want to control multiple cards")
        Else
            increaseIoOperations()
            Dim h As Integer = OpenDevice(CardAddress)

            If isBetween(h, 0, 3) Then
                activateCard(h)
                initializeConnection()
                connected_cards_array(h) = True
                Return True
            End If
        End If

        Return False
    End Function

    ''' <summary>
    ''' Disconnect the specified K8055 card
    ''' </summary>
    ''' <param name="cardNumber"></param>
    ''' <remarks></remarks>
    Public Sub disconnect(cardNumber As Integer)
        If cardNumber > -1 And cardNumber < MAX_NUMBER_OF_CARDS - 1 Then
            SetCurrentDevice(cardNumber)
            connected_cards_array(cardNumber) = False
            increaseIoOperations()
            CloseDevice()
        Else
            Throw New ArgumentException("You have specified an invalid card number!")
        End If
    End Sub

    ''' <summary>
    ''' Disconnect all specified K8055 cards
    ''' </summary>
    ''' <remarks></remarks>
    Public Sub disconnectAll()
        For i As Integer = 0 To UBound(connected_cards_array)
            SetCurrentDevice(i)
            increaseIoOperations()
            connected_cards_array(i) = False
            CloseDevice()
        Next
    End Sub

    ''' <summary>
    ''' Returns the software version of the K8055.DLL file from Velleman
    ''' </summary>
    ''' <returns>The version number</returns>
    ''' <remarks></remarks>
    Public Function getVersion() As Integer
        Return Version
    End Function
#End Region




#Region "Analog Output Channels"
    ''' <summary>
    ''' Set all analog output channels to the maximum value (255)
    ''' </summary>
    ''' <remarks></remarks>
    Public Sub setAllAnalogOutputsToMax()
        For i As Integer = 1 To MAX_ANALOG_OUTPUTS
            setAnalogOutputToMax(i)
        Next
    End Sub

    ''' <summary>
    ''' Clears all analog output channels (0)
    ''' </summary>
    ''' <remarks></remarks>
    Public Sub clearAllAnalogOutputs()
        For i As Integer = 1 To MAX_ANALOG_OUTPUTS
            clearAnalogOutput(i)
        Next
    End Sub

    ''' <summary>
    ''' Sets the specified output channel to the maximum value (255)
    ''' </summary>
    ''' <param name="channel">The channel to set (1-8)</param>
    ''' <remarks></remarks>
    Public Sub setAnalogOutputToMax(channel As Integer)
        Dim card, channel_of_card As Integer
        getChannelInformation(channel, channelType.ANALOG_OUTPUT, card, channel_of_card)
        If activateCard(card) Then
            increaseIoOperations()
            SetAnalogChannel(channel_of_card)
            analogOutputChannels(channel - 1) = MAX_DAC_STEPS
        End If
    End Sub

    ''' <summary>
    ''' Clears the specified analog output channel (0)
    ''' </summary>
    ''' <param name="channel">The channel to clear (1-8)</param>
    ''' <remarks></remarks>
    Public Sub clearAnalogOutput(channel As Integer)
        Dim card, channel_of_card As Integer
        getChannelInformation(channel, channelType.ANALOG_OUTPUT, card, channel_of_card)
        If activateCard(card) Then
            increaseIoOperations()
            ClearAnalogChannel(channel_of_card)
            analogOutputChannels(channel - 1) = 0
        End If
    End Sub

    ''' <summary>
    ''' Sets the specified analog output channel to the desired value
    ''' </summary>
    ''' <param name="channel">The channel to change (1-8)</param>
    ''' <param name="value">The desired output value (0-255)</param>
    ''' <remarks></remarks>
    Public Sub setAnalogOutputToValue(channel As Integer, value As Integer)
        If Not isBetween(value, 0, MAX_DAC_STEPS) Then
            Throw New ArgumentOutOfRangeException(
                String.Format("Incorrect analog value specified: Specify a value between 1 and {0}", MAX_DAC_STEPS))
        ElseIf _use_DAC_emulation And Not (value = 0 Or value = MAX_DAC_STEPS) Then
            Throw New Exception("You have enabled EMULATE_DAC mode: you can only set the DAC outputs to the minimum or maxium values")
        Else
            Dim card, channel_of_card As Integer
            getChannelInformation(channel, channelType.ANALOG_OUTPUT, card, channel_of_card)
            If activateCard(card) Then
                'setAnalog
                increaseIoOperations()
                OutputAnalogChannel(channel_of_card, value)
                analogOutputChannels(channel - 1) = value
            End If
        End If
    End Sub

    ''' <summary>
    ''' Sets the specified analog output channel to the desired value
    ''' </summary>
    ''' <param name="channel">The channel to change (1-8)</param>
    ''' <param name="value">The desired output value (0-255)</param>
    ''' <remarks></remarks>
    Public Sub setAnalogOutputToValue(channel As Integer, value As String)
        Dim integerValue As Integer
        If Not (Integer.TryParse(value, integerValue) AndAlso integerValue >= 0 AndAlso integerValue <= 255) Then
            Throw New ArgumentOutOfRangeException(
                String.Format("Incorrect analog value specified: Specify a value between 1 and {0}", MAX_DAC_STEPS))
        ElseIf _use_DAC_emulation And Not (integerValue = 0 Or integerValue = MAX_DAC_STEPS) Then
            Throw New Exception("You have enabled EMULATE_DAC mode: you can only set the DAC outputs to the minimum or maxium values")
        Else
            Dim card, channel_of_card As Integer
            getChannelInformation(channel, channelType.ANALOG_OUTPUT, card, channel_of_card)
            If activateCard(card) Then
                increaseIoOperations()
                OutputAnalogChannel(channel, integerValue)
                analogOutputChannels(channel - 1) = integerValue
            End If
        End If
    End Sub

    ''' <summary>
    ''' Sets all the analog output channels to the desired value
    ''' </summary>
    ''' <param name="value">The value for all channels (0-255)</param>
    ''' <remarks></remarks>
    Public Sub setAllAnalogOutputsToValue(value As Integer)
        If Not isBetween(value, 1, MAX_DAC_STEPS) Then
            Throw New ArgumentOutOfRangeException(
                String.Format("Incorrect analog value specified: Specify a value between 1 and {0}", MAX_DAC_STEPS))
        ElseIf _use_DAC_emulation And Not (value = 0 Or value = MAX_DAC_STEPS) Then
            Throw New Exception("You have enabled EMULATE_DAC mode: you can only set the DAC outputs to the minimum or maxium values")
        Else
            For i As Integer = 1 To MAX_DIGITAL_OUTPUTS
                setAnalogOutputToValue(i, value)
            Next
        End If
    End Sub

    ''' <summary>
    ''' Sets all the analog output channels to the desired value
    ''' </summary>
    ''' <param name="value">The string value for all channels (0-255)</param>
    ''' <remarks></remarks>
    Public Sub setAllAnalogOutputsToValue(value As String)
        Dim integerValue As Integer

        If Integer.TryParse(value, integerValue) AndAlso integerValue >= 0 AndAlso integerValue <= MAX_DAC_STEPS Then
            setAllAnalogOutputsToValue(integerValue)
        ElseIf _use_DAC_emulation And Not (integerValue = 0 Or integerValue = MAX_DAC_STEPS) Then
            Throw New Exception("You have enabled EMULATE_DAC mode: you can only set the DAC outputs to the minimum or maxium values")
        Else
            Throw New ArgumentOutOfRangeException(
                String.Format("Incorrect analog value specified: Specify a value between 1 and 255"))
        End If
    End Sub

    ''' <summary>
    ''' Gets the value of the specified analog output channel
    ''' </summary>
    ''' <param name="channel">The channel to poll (1-8)</param>
    ''' <returns>The current value (0-255)</returns>
    ''' <remarks></remarks>
    Public Function getAnalogOutput(channel As Integer) As Integer

        If channel > 0 And channel < analogOutputChannels.Count Then
            Return analogOutputChannels(channel - 1)
        Else
            Throw New ArgumentException(String.Format("You specified an invalid analog channel number. Specify a channel between 1 and {0}",
                                                      MAX_ANALOG_OUTPUTS))
        End If

    End Function
#End Region




#Region "Analog Input Channels"
    ''' <summary>
    ''' Read the value of the specified analog input channel
    ''' </summary>
    ''' <param name="channel">the channel to read (1-8)</param>
    ''' <returns>The value of the analog input (0-255)</returns>
    ''' <remarks></remarks>
    Public Function readAnalogInput(channel As Integer) As Integer
        If isBetween(channel, 1, MAX_ANALOG_INPUTS) Then
            If _timerEnabled And _cacheEnabled Then
                ' Cache mode enabled
                Return analogInputChannels(channel - 1)
            Else
                ' Cache mode disabled
                Dim card, channel_of_card As Integer
                getChannelInformation(channel, channelType.ANALOG_INPUT, card, channel_of_card)

                If activateCard(card) Then
                    increaseIoOperations()
                    Return ReadAnalogChannel(channel_of_card)
                End If
            End If
        Else
            Throw New ArgumentOutOfRangeException(
                String.Format("Incorrect analog channel specified: Specify a channel between 1 and {0}", MAX_ANALOG_INPUTS))
        End If

        'This can never happen
        Return 0
    End Function
#End Region





#Region "Class Initialisation & Event Generation"
    ''' <summary>
    ''' Returns the current object for the class (only one object can and should exist)
    ''' </summary>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Shared Function getInstance() As K8055
        Return currentInstance
    End Function

    ''' <summary>
    ''' The default constructor of the class. When calling this constructor there will be
    ''' no Events-generation.
    ''' </summary>
    ''' <remarks></remarks>
    Public Sub New()
        If currentInstance Is Nothing Then
            currentInstance = Me
            _timerEnabled = False
            initializeTimer()
        Else
            Throw New Exception("You are not allowed to create multiple objects of the K8055 class: object already exists. Obtain the current Object with the static K8055.getInstance() method")
        End If
    End Sub

    ''' <summary>
    ''' Private sub to run the built-in timer to check the card and eventually trigger Events.
    ''' </summary>
    ''' <remarks></remarks>
    Private Sub initializeTimer()
        readTimer = New System.Windows.Forms.Timer()
        readTimer.Enabled = True
        readTimer.Interval = _timerInterval
    End Sub

    ''' <summary>
    ''' readTimer_Cacher() will be execute at the specified interval. It will poll the K8055
    ''' and trigger an Event if necessary.
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    ''' <remarks></remarks>
    Private Sub readTimer_Cacher(ByVal sender As Object, ByVal e As System.EventArgs) Handles readTimer.Tick
        ' Only check the card if at least one K8055 card is connected.

        If _timerEnabled And countConnectedCards() > 0 Then
            ' Erase the initState indication.
            _initState = False

            Dim digitalChangeBuffer(MAX_NUMBER_OF_CARDS * MAX_DIGITAL_INPUTS - 1) As Boolean
            Dim analogChangeBuffer(MAX_NUMBER_OF_CARDS * MAX_ANALOG_INPUTS - 1) As Boolean
            Dim digitalData As Integer

            For curCard As Integer = 1 To MAX_NUMBER_OF_CARDS
                If connected_cards_array(curCard - 1) Then
                    ' The specified card is connected so we can include it in the test
                    If activateCard(curCard - 1) Then
                        If isOperationAllowed(curCard - 1, True) Then
                            digitalData = ReadAllDigital()
                            increaseIoOperations()

                            For curChannel As Integer = 1 To MAX_DIGITAL_INPUTS
                                Dim currentBit As Boolean =
                                    (digitalData And CInt(Math.Pow(2, curChannel - 1))) > 0

                                Dim curArrayIndex As Integer = ((curCard - 1) * MAX_DIGITAL_INPUTS + curChannel) - 1
                                digitalChangeBuffer(curArrayIndex) = digitalInputChannels(curArrayIndex) <> currentBit
                                digitalInputChannels(curArrayIndex) = currentBit
                            Next
                        End If

                        ' Read the AD channels in a single command to decrease the number of card readings
                        If isOperationAllowed(curCard - 1, False) Then
                            Dim adValues(MAX_ANALOG_INPUTS - 1) As Integer
                            ReadAllAnalog(adValues(0), adValues(1))
                            increaseIoOperations()

                            For curChannel As Integer = 1 To MAX_ANALOG_INPUTS
                                '   analogData = ReadAnalogChannel(curChannel)   ' 1-based reading!..
                                Dim curArrayIndex As Integer =
                                    ((curCard - 1) * MAX_ANALOG_INPUTS + curChannel) - 1

                                analogChangeBuffer(curArrayIndex) = adValues(curChannel - 1) <> analogInputChannels(curArrayIndex)
                                analogInputChannels(curArrayIndex) = adValues(curChannel - 1)
                            Next
                        End If
                    Else
                        ' An error occured while accessing the card
                        Debug.Print(String.Format("The card {0} seems to be disconnected from the computer", curCard))
                    End If
                End If
            Next

            ' Check if any Digital Input channels were changed
            For i As Integer = 1 To MAX_NUMBER_OF_CARDS * MAX_DIGITAL_INPUTS
                Dim cardid As Integer = (i - 1) \ MAX_DIGITAL_INPUTS
                Dim channelID As Integer = 1 + (i - 1) Mod MAX_DIGITAL_INPUTS

                If digitalChangeBuffer(i - 1) Then
                    RaiseEvent digitalInputChanged(cardid, channelID, i, digitalInputChannels(i - 1))

                    If digitalInputChannels(i - 1) Then
                        RaiseEvent digitalInputChangedOn(cardid, channelID, i, digitalInputChannels(i - 1))
                    Else
                        RaiseEvent digitalInputChangedOff(cardid, channelID, i, digitalInputChannels(i - 1))
                    End If
                End If
            Next

            ' Check if any Analog Input channels were changed
            For i As Integer = 1 To MAX_NUMBER_OF_CARDS * MAX_ANALOG_INPUTS
                Dim cardid As Integer = (i - 1) \ MAX_ANALOG_INPUTS
                Dim channelID As Integer = 1 + (i - 1) Mod MAX_ANALOG_INPUTS

                If analogChangeBuffer(i - 1) Then
                    RaiseEvent AnalogInputChanged(cardid, channelID, i, analogInputChannels(i - 1))
                End If
            Next

            ' Fire an additional event to indicate the processing is completed
            RaiseEvent readCycleCompleted()
        End If
    End Sub
#End Region





#Region "Aider Functions"
    ''' <summary>
    ''' Checks if the specified value lies between the allowed limits
    ''' </summary>
    ''' <param name="value">the value to check</param>
    ''' <param name="min">The minimum value</param>
    ''' <param name="max">The maximal value</param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Private Function isBetween(value As Integer, min As Integer, max As Integer) As Boolean
        Return (value >= min And value <= max)
    End Function

    ''' <summary>
    ''' Increase the number of I/O operations for statistical purposes
    ''' </summary>
    ''' <remarks></remarks>
    Private Sub increaseIoOperations()
        counted_IO_operations = CULng(counted_IO_operations + 1)
    End Sub

    ''' <summary>
    ''' Resizes the digitalOutputChannels() array, after changing the _use_DAC_emulation.
    ''' </summary>
    ''' <remarks></remarks>
    Private Sub convertDigitalOutputArray()
        clearAllAnalogOutputs()
        clearAllDigitalOutputs()

        If _use_DAC_emulation Then
            ReDim digitalOutputChannels(MAX_NUMBER_OF_CARDS * MAX_DIGITAL_OUTPUTS - 1)
        Else
            ReDim digitalOutputChannels(MAX_NUMBER_OF_CARDS * (MAX_DIGITAL_OUTPUTS + MAX_ANALOG_OUTPUTS) - 1)
        End If
    End Sub

    ''' <summary>
    ''' Wrapper function that should be called instead of the calling the Velleman routine SetAnalogChannel
    ''' directly. This is for compatibility with the _use_DAC_emulation.
    ''' </summary>
    ''' <param name="channel"></param>
    ''' <remarks></remarks>
    Private Sub setDigitalChannelWrapper(channel As Integer)
        increaseIoOperations()

        ' Directy call the Velleman procedures!
        If channel < 9 Then
            SetDigitalChannel(channel)
        Else
            SetAnalogChannel(channel - 8)
        End If
    End Sub

    ''' <summary>
    ''' Wrapper function that should be called instead of the calling the Velleman routine ClearAnalogChannel
    ''' directly. This is for compatibility with the _use_DAC_emulation.
    ''' </summary>
    ''' <param name="channel"></param>
    ''' <remarks></remarks>
    Private Sub clearDigitalChannelWrapper(channel As Integer)
        increaseIoOperations()

        ' Directy call the Velleman procedures!
        If channel < 9 Then
            ClearDigitalChannel(channel)
        Else
            ClearAnalogChannel(channel - 8)
        End If
    End Sub

    ''' <summary>
    ''' Checks if the specified option is enabled
    ''' </summary>
    ''' <param name="option">option to check</param>
    ''' <returns>True if the specified option is enabled</returns>
    ''' <remarks></remarks>
    Private Function isOptionEnabled([option] As EventConfiguration) As Boolean
        Return (currentEventConfiguration And [option]) > 0
    End Function

    ''' <summary>
    ''' Checks if a certain operation is allowed
    ''' </summary>
    ''' <param name="card">The current card</param>
    ''' <param name="isDigital">True to specify a digital operation, otherwise an analog operation</param>
    ''' <returns>True if the operation is allowed</returns>
    ''' <remarks></remarks>
    Private Function isOperationAllowed(card As Integer, isDigital As Boolean) As Boolean
        If isDigital And isOptionEnabled(EventConfiguration.ALL_DIGITAL) Then Return True
        If Not isDigital And isOptionEnabled(EventConfiguration.ALL_ANALOG) Then Return True

        Select Case card
            Case 0
                If isDigital Then
                    Return isOptionEnabled(EventConfiguration.CARD_0_DIGITAL)
                Else
                    Return isOptionEnabled(EventConfiguration.CARD_0_ANALOG)
                End If

            Case 1
                If isDigital Then
                    Return isOptionEnabled(EventConfiguration.CARD_1_DIGITAL)
                Else
                    Return isOptionEnabled(EventConfiguration.CARD_1_ANALOG)
                End If

            Case 2
                If isDigital Then
                    Return isOptionEnabled(EventConfiguration.CARD_2_DIGITAL)
                Else
                    Return isOptionEnabled(EventConfiguration.CARD_2_ANALOG)
                End If

            Case 3
                If isDigital Then
                    Return isOptionEnabled(EventConfiguration.CARD_3_DIGITAL)
                Else
                    Return isOptionEnabled(EventConfiguration.CARD_3_ANALOG)
                End If
        End Select

        ' All other cases are unknown
        Return False
    End Function
#End Region
End Class
